import { watch } from "node:fs/promises";
import chalk from "chalk";
import { CLIError, MuseError, loadFromBinding, resolveBindingPath } from "#lib";
import { renderPlaybillToHTML } from "#render/html";
import {
  type CLIArguments,
  getAbsoluteDirname,
  parseCLIArguments,
} from "#util";
import { version } from "../package.json" with { type: "json" };

const usage = `
${chalk.bold("muse")} - Compile and validate Playbills for Proscenium
${chalk.dim(`v${version}`)}

Usage:
  muse [/path/to/binding.yaml] <options>

Options:
  --check         Only load and check the current binding and resources, but do not compile
  --write         Write the output to a file. If not specified, outputs to stdout
  --outfile       Specify the output file path [default: playbill.json]
  --verbose, -v   Verbose output
  --minify, -m    Minify the output JSON
  --watch         Watch the directory for changes and recompile
  --help, -h      Show this help message
`.trim();

enum ExitCode {
  Success = 0,
  Error = 1,
}

async function processBinding({ inputFilePath, options }: CLIArguments) {
  // -- BINDING FILE -- //
  const bindingPath = await resolveBindingPath(inputFilePath);
  const binding = await loadFromBinding(bindingPath);

  // -- EXIT EARLY IF JUST CHECKING VALIDATION --//
  if (options.check) {
    console.log(chalk.green("Playbill validated successfully"));
    return ExitCode.Success;
  }

  // -- SERIALIZE USING RENDERER --//
  let serializedPlaybill = "";

  switch (options.renderer) {
    case "json":
      serializedPlaybill = binding.playbill.serialize();
      break;
    case "html":
      serializedPlaybill = await renderPlaybillToHTML(binding.playbill, {
        styles: binding.styles,
      });
      break;
    default:
      throw new CLIError(`Unknown renderer: ${options.renderer}`);
  }

  // -- WRITE TO DISK OR STDOUT --//
  if (options.write) {
    await Bun.write(options.outfile, serializedPlaybill);

    return ExitCode.Success;
  }

  console.log(serializedPlaybill);
  return ExitCode.Success;
}

async function main(): Promise<number> {
  const cliArguments = parseCLIArguments(Bun.argv.slice(2));
  const { options } = cliArguments;

  // -- HELP TEXT -- //
  if (options.help) {
    console.log(usage);
    return ExitCode.Success;
  }

  console.log(`Processing ${cliArguments.inputFilePath}`);
  let lastProcessResult = await processBinding(cliArguments);

  if (options.watch) {
    const watchDir = getAbsoluteDirname(cliArguments.inputFilePath);

    console.log(`Watching ${watchDir} for changes`);

    const watcher = watch(watchDir, {
      recursive: true,
    });

    for await (const event of watcher) {
      console.log(`Detected ${event.eventType} on ${event.filename}`);

      try {
        lastProcessResult = await processBinding(cliArguments);

        if (lastProcessResult === ExitCode.Error) {
          console.error(`Error processing ${event.filename}`);
        } else {
          console.log("Reprocessed changes");
        }
      } catch (error) {
        console.error(`Error processing ${event.filename}`);
        console.error(error);
      }
    }
    return ExitCode.Success;
  }

  return ExitCode.Success;
}

try {
  const exitCode = await main();
  process.exit(exitCode);
} catch (error) {
  if (error instanceof MuseError) {
    console.error(chalk.red(error.fmt()));
  } else {
    console.error("An unexpected error occurred");
    console.error(error);
  }

  process.exit(1);
}
