import path from "node:path";
import {
  type Ability,
  type Blueprint,
  type Item,
  type Method,
  Playbill,
  type Resource,
  type Rule,
  type Species,
} from "@proscenium/playbill";
import { Glob } from "bun";
import z from "zod";
import { loadYAMLFileOrFail } from "#util";
import { ComponentFile } from "./component-file";
import { FileNotFoundError } from "./errors";

const BindingSchema = z.object({
  extend: z.string().optional(),

  id: z.string().default("playbill"),
  name: z.string().default("Unnamed playbill"),
  author: z.string().optional(),
  description: z.string().optional(),
  version: z.string().default("0.0.0"),

  files: z.array(z.string()).default(["**/*.yaml", "**/*.md"]),
  styles: z.array(z.string()).optional(),
});

type Binding = z.infer<typeof BindingSchema>;

export interface BoundPlaybill {
  _raw: Binding;

  // File information
  bindingFilePath: string;
  bindingFileDirname: string;
  includedFiles: string[];

  // Binding properties
  id: string;
  name: string;
  author: string;
  version: string;
  description: string;

  styles?: string;

  playbill: Playbill;
}

/**
 * Given a path, find the binding.yaml file
 * If the path is to a directory, check for a binding.yaml inside
 * If the path is to a file, check if it exists and is a binding.yaml
 * Otherwise throw an error
 *
 * @param bindingPath - The path to the binding file
 * @returns The absolute resolved path to the binding file
 */
export async function resolveBindingPath(bindingPath: string): Promise<string> {
  // If the path does not specify a filename, use binding.yaml
  const inputLocation = Bun.file(bindingPath);

  // If it is a directory, try again seeking a binding.yaml inside
  const stat = await inputLocation.stat();
  if (stat.isDirectory()) {
    return resolveBindingPath(path.resolve(bindingPath, "binding.yaml"));
  }

  // If it doesnt exist, bail
  if (!(await inputLocation.exists())) {
    throw new FileNotFoundError(bindingPath);
  }

  // Resolve the path
  return path.resolve(bindingPath);
}

export async function loadFromBinding(
  bindingPath: string,
): Promise<BoundPlaybill> {
  const resolvedBindingPath = await resolveBindingPath(bindingPath);
  const yamlContent = await loadYAMLFileOrFail(resolvedBindingPath);
  const binding = BindingSchema.parse(yamlContent);
  const fileGlobs = binding.files;
  const bindingFileDirname = path.dirname(resolvedBindingPath);

  const playbill = new Playbill();

  // If this is extending another binding, load that first
  if (binding.extend) {
    const pathFromHere = path.resolve(bindingFileDirname, binding.extend);
    const extendBindingPath = await resolveBindingPath(pathFromHere);
    const loadedBinding = await loadFromBinding(extendBindingPath);
    playbill.extend(loadedBinding.playbill);
  }

  // Load any specified stylesheets
  let styles = "";
  if (binding.styles && binding.styles.length > 0) {
    for (const style of binding.styles) {
      const cssFilePath = path.resolve(bindingFileDirname, style);
      const cssFile = Bun.file(cssFilePath);
      const cssFileExists = await cssFile.exists();

      if (cssFileExists) {
        styles += `${await cssFile.text()}\n`;
      } else {
        throw new FileNotFoundError(cssFilePath);
      }
    }
  }

  // Scan for all files matching the globs
  const allFilePaths: string[] = [];

  for (const thisGlob of fileGlobs) {
    const glob = new Glob(thisGlob);

    const results = glob.scanSync({
      cwd: bindingFileDirname,
      absolute: true,
      followSymlinks: true,
      onlyFiles: true,
    });

    allFilePaths.push(...results);
  }

  // Exclude the binding.yaml file
  const filePathsWithoutBindingFile = allFilePaths.filter((filePath) => {
    return filePath !== resolvedBindingPath;
  });

  // Load discovered component files
  const componentFiles: ComponentFile[] = [];
  const fileLoadPromises: Promise<void>[] = [];

  for (const filepath of filePathsWithoutBindingFile) {
    const componentFile = new ComponentFile(filepath);
    componentFiles.push(componentFile);
    fileLoadPromises.push(componentFile.load());
  }

  await Promise.all(fileLoadPromises);

  // Decorate the playbill with the binding metadata
  playbill.name = binding.name;
  playbill.author = binding.author ?? "Anonymous";
  playbill.version = binding.version;
  playbill.description = binding.description ?? "";

  const abilities: Ability[] = [];
  const bluprints: Blueprint[] = [];
  const glossaries: Record<string, string[]>[] = [];
  const methods: Method[] = [];
  const species: Species[] = [];
  const items: Item[] = [];
  const rules: Rule[] = [];
  const resources: Resource[] = [];

  // Aggregate all components before adding to playbill to allow for
  // ensuring components are added in a valid order
  for (const file of componentFiles) {
    // Load components from the file into the playbill
    for (const component of file.components) {
      switch (component.type) {
        case "ability":
          abilities.push(component.component);
          break;
        case "blueprint":
          bluprints.push(component.component);
          break;
        case "glossary":
          glossaries.push(component.component);
          break;
        case "method":
          methods.push(component.component);
          break;
        case "species":
          species.push(component.component);
          break;
        case "item":
          items.push(component.component);
          break;
        case "rule":
          rules.push(component.component);
          break;
        case "resource":
          resources.push(component.component);
          break;
        default:
          throw new Error("Unknown component type");
      }
    }
  }

  // Add resources first
  for (const resource of resources) {
    playbill.addResource(resource);
  }
  // Add abilities before methods, species, and blueprints
  for (const ability of abilities) {
    playbill.addAbility(ability);
  }

  // Add species before blueprints
  for (const specie of species) {
    playbill.addSpecies(specie);
  }

  // Add methods after abilities
  for (const method of methods) {
    playbill.addMethod(method);
  }

  // Add items
  for (const item of items) {
    playbill.addItem(item);
  }

  // Add blueprints after methods, species, items, and abilities
  for (const blueprint of bluprints) {
    playbill.addBlueprint(blueprint);
  }

  // Add all definitions
  for (const glossary of glossaries) {
    for (const [term, definitions] of Object.entries(glossary)) {
      for (const definition of definitions) {
        playbill.addDefinition(term, definition);
      }
    }
  }

  return {
    _raw: binding,
    bindingFilePath: bindingPath,
    bindingFileDirname,
    includedFiles: filePathsWithoutBindingFile,

    id: binding.id,
    name: binding.name,
    author: binding.author ?? "Anonymous",
    description: binding.description ?? "",
    version: binding.version,

    styles: styles.length > 0 ? styles : undefined,

    playbill,
  };
}
