import type {
  ImageResource,
  Resource,
  TableResource,
  TextResource,
} from "@proscenium/playbill";
import { html, renderMarkdown } from "#lib";
import { Section } from "./base/section";

async function TextResourceSection(resource: TextResource): Promise<string> {
  const renderedMarkdown = await renderMarkdown(resource.description);

  const info = resource.categories
    .map((category) => {
      return html`<li>${category}</li>`;
    })
    .join("\n");

  return Section({
    type: "resource",
    componentId: resource.id,
    title: resource.name,
    content: renderedMarkdown,
    info: `<ul class="resource-categories">${info}</ul>`,
  });
}

async function ImageResourceSection(resource: ImageResource): Promise<string> {
  const renderedMarkdown = await renderMarkdown(resource.description);

  const info = resource.categories
    .map((category) => {
      return html`<li>${category}</li>`;
    })
    .join("\n");

  const fig = html`
    <figure class="resource-image">
      <img src="${resource.url}" alt="${resource.name}" />
      <figcaption>${renderedMarkdown}</figcaption>
    </figure>
  `;

  return Section({
    type: "resource",
    componentId: resource.id,
    title: resource.name,
    content: fig,
    info: `<ul class="resource-categories">${info}</ul>`,
  });
}

async function TableResourceSection(resource: TableResource): Promise<string> {
  const renderedMarkdown = await renderMarkdown(resource.description);

  const info = resource.categories
    .map((category) => {
      return html`<li>${category}</li>`;
    })
    .join("\n");

  const tableData = resource.data;

  const [headerRow, ...rows] = tableData;

  let tableHTML = html`<table class="resource-table"><thead><tr>`;

  for (const header of headerRow) {
    tableHTML += html`<th>${await renderMarkdown(header)}</th>`;
  }

  tableHTML += html`</tr></thead><tbody>`;

  for (const row of rows) {
    tableHTML += html`<tr>`;
    for (const cell of row) {
      tableHTML += html`<td>${await renderMarkdown(cell)}</td>`;
    }
    tableHTML += html`</tr>`;
  }

  tableHTML += html`</tbody></table>`;

  return Section({
    type: "resource",
    componentId: resource.id,
    title: resource.name,
    content: renderedMarkdown + tableHTML,
    info: `<ul class="resource-categories">${info}</ul>`,
  });
}

export async function ResourceSection(resource: Resource): Promise<string> {
  switch (resource.type) {
    case "text":
      return TextResourceSection(resource);
    case "image":
      return ImageResourceSection(resource);
    case "table":
      return TableResourceSection(resource);
  }
}
